#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytorch_lightning as pl


class LitModel(pl.LightningModule):
    def __init__(self, model, criteria=None, optim_config=None):
        super().__init__()
        self.save_hyperparameters()
        self.model = model
        self.criteria = criteria

        if optim_config is not None:
            optim_config = dict(optim_config)
        self.optim_config = optim_config

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self.model(x)
        loss = self.criteria(logits, y)
        return loss

    def training_step_end(self, batch_parts):
        loss = batch_parts
        self.log_dict({"training/loss": loss})
        # TODO gradient hist

    @property
    def val_metrics(self):
        raise NotImplementedError

    @property
    def test_metrics(self):
        raise NotImplementedError

    def training_epoch_end(self, training_step_outputs):
        return super().training_epoch_end(training_step_outputs)

    def validation_step(self, batch, batch_idx):
        x, y = batch
        outputs = self.model(x)
        loss = self.criteria(outputs, y)
        return {"loss": loss, "outputs": outputs, "y": y}

    def validation_step_end(self, batch_parts):
        loss = batch_parts["loss"]
        y = batch_parts["y"]
        outputs = batch_parts["outputs"]
        metrics = {}
        metrics = self.val_metrics(outputs, y)
        metrics.update({"val/loss": loss})
        self.log_dict(metrics)

    def validation_epoch_end(self, validation_step_outputs):
        return super().validation_epoch_end(validation_step_outputs)

    def test_step(self, batch, batch_idx):
        return self.validation_step(batch, batch_idx)

    def test_step_end(self, batch_parts):
        loss = batch_parts["loss"]
        y = batch_parts["y"]
        outputs = batch_parts["outputs"]
        metrics = {}
        metrics = self.test_metrics(outputs, y)
        metrics.update({"test/loss": loss})
        self.log_dict(metrics)

    def test_epoch_end(self, test_step_outputs):
        return super().test_epoch_end(test_step_outputs)

    def configure_optimizers(self):
        if self.optim_config is None:
            raise NotImplementedError

        else:
            return self.optim_config



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch.nn as nn
from transformers import AutoConfig, AutoModelForSequenceClassification


class BertSequenceClassification(nn.Module):
    def __init__(self, bert_model_dir: str, num_labels: int, **kwargs):
        """
        Keyword Arguments:
        bert_model_dir:str --
        num_labels:int     -- (default 2)
        **kwargs           --
        """
        super(BertSequenceClassification, self).__init__()
        kwargs.update(num_labels=num_labels)
        config = AutoConfig.from_pretrained(bert_model_dir, **kwargs)

        # TODO: use BertModel.from_pretrained instead
        self.bert = AutoModelForSequenceClassification.from_pretrained(
            bert_model_dir, config=config
        )

    def forward(self, x):
        input_ids, attention_mask, token_type_ids = x
        output = self.bert(input_ids, attention_mask, token_type_ids)
        logits = output[0]
        return logits



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, List

from omeganlp.core.mixins import HyperparametersMixin
from omeganlp.core.io import read_vocab_txt
import os


class Vocabulary(HyperparametersMixin):
    def __init__(
        self,
        vocab,
        padding_idx=0,
        unk_idx=1,
        pad="[PAD]",
        unk="[UNK]",
        oov_subword=True,
    ):
        super().__init__()
        if isinstance(vocab, str) and os.path.exists(vocab):
            vocab = read_vocab_txt(vocab)
        if pad in vocab and vocab.index(pad) != padding_idx:
            vocab.pop(vocab.index(pad))
            vocab.insert(padding_idx, pad)
        if unk in vocab and vocab.index(unk) != unk_idx:
            vocab.pop(vocab.index(unk))
            vocab.insert(unk_idx, unk)

        self.save_hyperparameters()

        self.set_hparams_attributes()

        self._stoi = {word: i for i, word in enumerate(self.vocab)}
        self._itos = {i: word for i, word in enumerate(self.vocab)}

    def stoi(self, word):
        return self._stoi.get(word, self.unk_idx)

    def itos(self, index):
        return self._itos.get(index, self.unk)

    def __len__(self):
        return len(self.vocab)

    @classmethod
    def build_from_corpus(cls, corpus: Iterable[str], tokenizer, **kwargs):
        vocab = []
        for c in corpus:
            vocab += tokenizer(c)

        vocab = list(set(vocab))
        return cls(vocab=vocab, **kwargs)

    def encode(self, text: str):
        idx = self.stoi(text)
        if self.oov_subword and idx == self.unk_idx and len(text) > 1:
            text = list(text)
            idx = [self.stoi(t) for t in text]

        else:
            idx = [idx]

        return idx

    def __call__(self, sequence: List[str], max_length) -> List[int]:
        encoded = []
        for s in sequence:
            encoded += self.encode(s)
        seq_len = len(encoded)
        if seq_len <= max_length:
            encoded += [self.padding_idx] * (max_length - seq_len)
        else:
            encoded = encoded[:max_length]
        return encoded



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pytorch_lightning.core.mixins import (
    HyperparametersMixin as PL_HyperparametersMixin,
)


class HyperparametersMixin(PL_HyperparametersMixin):
    def __init__(self):
        super().__init__()

    def update_hparams(self):
        for k in self.hparams.keys():
            if hasattr(self, k):
                v = getattr(self, k)
                self.hparams.update({k: v})

    def set_hparams_attributes(self):
        for k, v in self.hparams.items():
            setattr(self, k, v)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib
import inspect
import json
import os
import shutil
import tarfile
from typing import Any, Callable, Dict, Iterable, List, MutableMapping, Optional, Union

from omegaconf import DictConfig, OmegaConf


def _read_json_per_line(file_path, filter_, ignore_exception):
    with open(file_path) as f:
        for line in f:
            try:
                data = json.loads(line)
                if filter_ is None or filter_(data):
                    yield data
            except Exception as e:
                if not ignore_exception:
                    raise e


def read_json(
    file_path: str,
    per_line_mode: bool = True,
    filter_: Optional[Callable[[Dict], bool]] = None,
    ignore_exception: bool = False,
) -> Union[Iterable[Dict], Dict]:

    if per_line_mode:
        return _read_json_per_line(file_path, filter_, ignore_exception)
    else:
        with open(file_path) as f:
            try:
                return json.load(f)
            except Exception as e:
                if not ignore_exception:
                    raise e


def dump_json(
    objs: Union[Iterable[Any], Any],
    file_path: str,
    per_line_mode: bool = True,
    ensure_ascii: bool = False,
) -> None:
    with open(file_path, "w") as f:
        if per_line_mode:
            for obj in objs:
                json.dump(obj, f, ensure_ascii=ensure_ascii)
                f.write("\n")
        else:
            json.dump(objs, f, ensure_ascii=ensure_ascii)


def read_vocab_txt(file_path: str) -> List[str]:
    results = []
    with open(file_path) as f:
        for line in f:
            results += [line.strip()]
    return results


def dump_vocab_txt(vocab: Iterable[Union[str, Any]], file_path: str) -> None:
    with open(file_path, "w") as f:
        for v in vocab:
            f.write(v)
            f.write("\n")


def tgz(source_dir: str, output_path: Optional[str] = None) -> str:
    basename = os.path.basename(source_dir)
    if output_path is None:
        output_path = basename + ".tgz"
    with tarfile.open(output_path, "w:gz") as tar:
        tar.add(source_dir, arcname=basename)
    return os.path.abspath(output_path)


def untgz(source_path: str, target_dir: str = ".") -> None:
    with tarfile.open(source_path, "r") as tar:
        tar.extractall(path=target_dir)


def is_empty(path: str) -> Optional[bool]:
    if os.path.exist(path):
        files = os.listdir(path)
        return len(files) <= 0


def clean_dir(path: str) -> None:
    if os.path.exists(path):
        shutil.rmtree(path)
    os.makedirs(path, exist_ok=True)


class OmegaDataDir:
    root = os.environ.get("OMEGA_DATA")
    if root is None:
        default_path = os.path.expanduser("~/.omega")
        import warnings

        warnings.warn(
            "No Environment Variable:OMEGA_DATA Detected! Automatically set OMAGA_DATA by default at {}".format(
                default_path
            )
        )
        del warnings
        os.makedirs(default_path, exist_ok=True)
        root = default_path

    pretrained = os.path.join(root, "pretrained")
    datasets = os.path.join(root, "datasets")
    embeddings = os.path.join(root, "pretrained", "embeddings")
    sota = os.path.join(root, "sota")
    extras = os.path.join(root, "extras")

    @staticmethod
    def parse(path: str):
        if isinstance(path, str) and path.startswith("omega://"):
            return os.path.join(OmegaDataDir.root, path[8:])
        else:
            return path

    @staticmethod
    def parse_config(config: MutableMapping):
        for k, v in config.items():
            if isinstance(v, str):
                _v = OmegaDataDir.parse(v)
            elif isinstance(v, dict):
                _v = OmegaDataDir.parse_mapping(v)
            else:
                _v = v
            config.update({k: _v})
        return config


def check_path(data_dir, name, ignore_exception=False):
    names = os.listdir(data_dir)
    names = [n for n in names if n.split(".")[0] == name]
    if len(names) == 0:
        if not ignore_exception:
            raise Exception(f"No file named {name} in {data_dir}")
    else:
        return os.path.join(data_dir, names[0])


def dump_config(config, config_path):
    assert config_path.endswith(".yaml")
    with open(config_path, "w") as f:
        OmegaConf.save(DictConfig(config), f)


def read_config(config_path):
    assert config_path.endswith(".yaml")
    with open(config_path) as f:
        config = OmegaConf.load(f)
    return config


def find_class(module_class_path):
    module_name = ".".join(module_class_path.split(".")[:-1])
    class_name = module_class_path.split(".")[-1]
    module = importlib.import_module(module_name)
    target_cls = getattr(module, class_name)
    return target_cls


def get_class_fullname(o: Any):
    module = o.__class__.__module__
    name = o.__class__.__name__
    return ".".join([module, name])


def incept_function_args(func):
    args = inspect.getfullargspec(func)
    args = args.args
    return [a for a in args if a != "self"]


def get_attr_method_from_module(module_name, attr_method_name):
    module = importlib.import_module(module_name)
    attr_method = getattr(module, attr_method_name)
    return attr_method


def instantiate_interface(target: str):
    module, *attrs = target.split(".")
    module = importlib.import_module(module)

    interface = module
    for attr in attrs:
        interface = getattr(interface, attr)
    return interface



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchmetrics import Accuracy, MetricCollection

from omeganlp.litmodel import LitModel


class SequenceClassificationModel(LitModel):
    @property
    def val_metrics(self):
        return MetricCollection({"val/acc": Accuracy(top_k=1)})

    @property
    def test_metrics(self):
        return MetricCollection({"test/acc": Accuracy(top_k=1)})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
import torch.nn.functional as F


class TextCNN(nn.Module):
    def __init__(
        self,
        num_labels,
        vocab_size,
        embedding_dim=100,
        embedding=None,
        freeze_embedding=False,
        kernel_sizes=[2, 3, 4, 5],
        filter_num=2,
        dropout=0.5,
        **kwargs
    ):
        super(TextCNN, self).__init__()

        if embedding is not None:
            self.embedding = nn.Embedding.from_pretrained(embedding, freeze=freeze_embedding, **kwargs)
        else:
            # embedding initiailzation is Gaussians by default
            self.embedding = nn.Embedding(vocab_size, embedding_dim, **kwargs)

        self.convs = nn.ModuleList(
            [
                nn.Conv2d(
                    in_channels=1,
                    out_channels=filter_num,
                    kernel_size=(k, embedding_dim),
                )
                for k in kernel_sizes
            ]
        )

        self.dropout = nn.Dropout(dropout)
        self.fc = nn.Linear(in_features=len(kernel_sizes) * filter_num, out_features=num_labels)

    def embedding_encoding(self, x):
        return self.embedding(x)

    def embedding_modeling(self, embedding):
        x = embedding.unsqueeze(1)
        x = [F.relu(conv(x)).squeeze(3) for conv in self.convs]
        x = [F.max_pool1d(item, item.size(2)).squeeze(2) for item in x]
        x = torch.cat(x, 1)
        x = self.dropout(x)
        logits = self.fc(x)
        return logits

    def forward(self, x):
        embedding = self.embedding_encoding(x)
        return self.embedding_modeling(embedding)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pytorch_lightning.callbacks import Callback


class ExceptionCallback(Callback):
    def on_exception(trainer, litmodel, exception):
        print("saving checkpoints by Exception")
        # TODO:save checkpoint
        raise exception



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

import hydra
from omegaconf import DictConfig

from omeganlp.core.io import find_class


class Task:
    def __init__(self, trainer, litmodel, datamodule, processor=None):
        self.trainer = trainer
        self.litmodel = litmodel
        self.datamodule = datamodule

        if datamodule is not None:
            self.processor = datamodule.processor
        else:
            self.processor = processor

        self.root_dir = self.trainer.default_root_dir
        self.model = litmodel.model

    def train(self):
        return self.trainer.fit(self.litmodel, self.datamodule)

    def validate(self):
        return self.trainer.validate(self.litmodel, self.datamodule)

    def test(self):
        return self.trainer.test(self.litmodel, self.datamodule)

    def infer(self, examples):
        pass

    def save_inferer(
        self, inferer_dir: str = "./inferer", to_onnx: bool = False, compress=False
    ):
        inferer_dir = os.path.join(self.root_dir, inferer_dir)
        os.makedirs(inferer_dir, exist_ok=True)

        model_save_dir = os.path.join(inferer_dir, "model")
        os.makedirs(model_save_dir, exist_ok=True)

        processor_save_dir = os.path.join(self.root_dir, "processor")
        os.makedirs(processor_save_dir, exist_ok=True)

        self.processor.save(processor_save_dir)
        # TODO: save best model
        if to_onnx:
            # TODO:try to trans to onnx model
            pass
        # TODO: save litmodel
        # TODO: save task

    def load_inferer(self, inferer_dir: str = "./inferer"):
        pass

    def load_best_model(self):
        pass

    @staticmethod
    def update_model(datamodule):
        return {}

    @staticmethod
    def load(config: DictConfig):
        task_cls = find_class(config.task._target_)

        datamodule = hydra.utils.instantiate(config.datamodule)
        datamodule.prepare_data()

        num_training_steps = config.trainer.max_epochs * datamodule.num_step("train")

        add_model_params = task_cls.update_model(datamodule)

        model = hydra.utils.instantiate(config.litmodel.model, **add_model_params)

        optim = hydra.utils.instantiate(
            config.optim, model=model, num_training_steps=num_training_steps
        )

        litmodel = hydra.utils.instantiate(
            config.litmodel, model=model, optim_config=optim.config
        )

        # cbs = [config.callbacks.model_checkpoint, config.callbacks.early_stopping,
        #        config.callbacks.exception]
        # TODO: exception save checkpoint
        cbs = [config.callbacks.model_checkpoint, config.callbacks.early_stopping]
        callbacks = [hydra.utils.instantiate(cb) for cb in cbs]

        trainer = hydra.utils.instantiate(config.trainer, callbacks=callbacks)
        task = hydra.utils.instantiate(config.task, trainer, litmodel, datamodule)

        return task


class SequenceClassificationTask(Task):
    @staticmethod
    def update_model(datamodule):
        processor = datamodule.processor
        model_params = {"num_labels": processor.num_labels}

        if hasattr(processor, "vocabulary"):
            model_params.update({"vocab_size": len(processor.vocabulary)})

        return model_params



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
from typing import Optional

import numpy as np
from pytorch_lightning import LightningDataModule

from omeganlp.core.io import check_path
from omeganlp.data.processor import BaseDataProcessor


class DataType:
    TRAIN = "train"
    DEV = "dev"
    TEST = "test"


class DataModule(LightningDataModule):
    def __init__(
        self,
        processor: BaseDataProcessor,
        data_root: str,
        save_processed_data: bool = False,
        batch_size=None,
        train_batch_size=64,
        val_batch_size=64,
        test_batch_size=64,
        **kwargs
    ):
        super().__init__()

        if batch_size is not None:
            train_batch_size = batch_size
            val_batch_size = batch_size
            test_batch_size = batch_size

        # self.save_hyperparameters()

        self.processor = processor
        self.data_root = data_root

        self.train_batch_size = train_batch_size
        self.val_batch_size = val_batch_size
        self.test_batch_size = test_batch_size

        self.trainset = []
        self.valset = []
        self.testset = []

        self.save_processed_data = save_processed_data
        self.dataloader_kwargs = kwargs

    def get_dataset(self, dataset_path: Optional[str] = None, mode=DataType.TRAIN):
        """ Agent of processor.get_dataset """

        if dataset_path is None and self.data_root is not None:
            dataset_path = check_path(self.data_root, mode, ignore_exception=True)

        if isinstance(dataset_path, str) and os.path.exists(dataset_path):
            if dataset_path.endswith("pt"):
                dataset = self.processor.load_processed(dataset_path)
            else:
                is_train = True
                dataset = self.processor.get_dataset(dataset_path, is_train=is_train)

        return dataset

    def dataset2dataloader(self, dataset, batch_size):
        """ Agent of processor.dataset2dataloader """
        return self.processor.dataset2dataloader(
            dataset, batch_size, **self.dataloader_kwargs
        )

    def num_step(self, mode=DataType.TRAIN):
        def _num_step(dataset, batch_size):
            return int(np.ceil(len(dataset) / batch_size))

        if mode == DataType.TRAIN:
            return _num_step(self.trainset, self.train_batch_size)

        elif mode == DataType.DEV:
            return _num_step(self.valset, self.val_batch_size)

        elif mode == DataType.TEST:
            return _num_step(self.testset, self.test_batch_size)

    def prepare_data(self):
        # TODO: may move to setup method
        if check_path(self.data_root, DataType.TRAIN):
            self.trainset = self.get_dataset(mode=DataType.TRAIN)

        if check_path(self.data_root, DataType.DEV):
            self.valset = self.get_dataset(mode=DataType.DEV)

        if check_path(self.data_root, DataType.TEST):
            self.testset = self.get_dataset(mode=DataType.TEST)

        if self.save_processed_data:
            data_dir = os.path.join(self.data_root, ".processed")
            os.makedirs(data_dir, exist_ok=True)

            self.processor.save_processed(
                self.trainset, os.path.join(data_dir, "train.pt")
            )
            self.processor.save_processed(self.valset, os.path.join(data_dir, "dev.pt"))
            self.processor.save_processed(
                self.testset, os.path.join(data_dir, "test.pt")
            )

    def setup(self, stage):
        pass

    def train_dataloader(self):
        return self.dataset2dataloader(self.trainset, batch_size=self.train_batch_size)

    def val_dataloader(self):
        return self.dataset2dataloader(self.valset, batch_size=self.val_batch_size)

    def test_dataloader(self):
        return self.dataset2dataloader(self.testset, batch_size=self.test_batch_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from transformers import AdamW, get_linear_schedule_with_warmup
from omeganlp.core.io import instantiate_interface


def default_pretrained_optimizer_and_scheduler(
    model, num_training_steps, num_warmup_steps=0.01, weight_decay=0.01, lr=1e-5
):
    model_parameters = list(model.named_parameters())
    no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]

    optimizer_grouped_parameters = [
        {
            "params": [
                p for n, p in model_parameters if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": weight_decay,
        },
        {
            "params": [
                p for n, p in model_parameters if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]

    optimizer = AdamW(optimizer_grouped_parameters, lr=lr)

    if num_warmup_steps < 1:
        num_warmup_steps = int(num_training_steps * num_warmup_steps)

    lr_scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=num_warmup_steps,
        num_training_steps=num_training_steps,
    )

    return optimizer, lr_scheduler


class OptimComponents:
    def __init__(
        self, optim_fn, model, optimizer_config=None, lr_scheduler_config=None, **kwargs
    ):
        # TODO: make optimizer_config do
        if isinstance(optim_fn, str):
            optim_fn = instantiate_interface(optim_fn)
        optimizer, lr_scheduler = optim_fn(model, **kwargs)
        config = {"optimizer": optimizer}

        if lr_scheduler is not None:
            if lr_scheduler_config is not None:
                lr_scheduler_config.update({"scheduler": lr_scheduler})
            else:
                lr_scheduler_config = lr_scheduler
            config.update({"lr_scheduler": lr_scheduler_config})

        self._config = config

    @property
    def config(self):
        return self._config



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from abc import ABCMeta, abstractmethod
from typing import Any

import torch


class BaseDataProcessor(metaclass=ABCMeta):
    @abstractmethod
    def processing(self, example: Any, is_train: bool = False):
        """
        example --> features
        Process an example(a single sample) to your specified encoded features

        Keyword Arguments:
        example:Any   -- your data example
        is_train:bool -- (default False) whether the example contains 'y'
        """

    @abstractmethod
    def read(self, data_dir: str, **kwargs):
        """
        Keyword Arguments:
        data_dir:str -- raw data dir
        """

    @abstractmethod
    def get_dataset(self, data_dir: str, is_train: bool = False, **kwargs):
        """
        Keyword Arguments:
        data_dir:str -- raw data dir
        """

    @abstractmethod
    def save(self, **kwargs):
        """
        save the metadata for your processor
        or pickle the processor
        """

    @classmethod
    @abstractmethod
    def load(cls, **kwargs):
        """
        call load classmethod to load saved metadata and initialize processor
        or unpickle the processor
        """

    def data2dataset(self, datas, is_train: bool = False, **kwargs):
        raise NotImplementedError

    @staticmethod
    def dataset2dataloader(dataset, batch_size, **kwargs):
        raise NotImplementedError

    def update_params(self, **kwargs):
        raise NotImplementedError

    def decode(self, **kwargs):
        raise NotImplementedError

    @staticmethod
    def load_processed(data_dir):
        return torch.load(data_dir)

    @staticmethod
    def save_processed(data, data_dir):
        dirname = os.path.dirname(data_dir)
        os.makedirs(dirname, exist_ok=True)
        torch.save(data, data_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from dataclasses import dataclass
from typing import Dict, Iterable, List, Optional, Union

import torch
from torch.utils.data import Dataset, TensorDataset
from torch.utils.data.dataloader import DataLoader, default_collate
from tqdm import tqdm
from transformers import AutoTokenizer, BertTokenizer

from omeganlp.core.io import (
    dump_config,
    dump_json,
    dump_vocab_txt,
    get_class_fullname,
    instantiate_interface,
    read_config,
    read_json,
)
from omeganlp.core.mixins import HyperparametersMixin
from omeganlp.core.vocabulary import Vocabulary

from . import BaseDataProcessor


@dataclass
class Example:
    text_a: str
    text_b: str
    label: Union[str, int, List[Union[str, int]]]
    guid: str = ""


class SequenceClassificationDataset(Dataset):
    def __init__(self, x, y=None):
        self.x = x
        self.y = y

    def __len__(self):
        return len(self.x)

    def __getitem__(self, index):
        if self.y is None:
            return self.x[index]
        else:
            return self.x[index], self.y[index]


class BaseSequenceClassificationProcessor(BaseDataProcessor, HyperparametersMixin):
    def __init__(
        self,
        labels: Union[List[Union[int, str]], str],
        tokenizer_dir: str,
        max_length: int,
        key_text_a: str = "text_a",
        key_text_b: str = "text_b",
        key_label="label",
        key_guid: str = "_id",
        **kwargs
    ):
        super().__init__()

        if isinstance(labels, str):
            assert os.path.exists(labels)
            labels = list(read_json(labels))
        assert len(set(labels)) > 1 and len(labels) == len(set(labels))

        self.save_hyperparameters()

        self.max_length = max_length
        self.labels = labels
        self.tokenizer_dir = tokenizer_dir
        self.key_text_a = key_text_a
        self.key_text_b = key_text_b
        self.key_label = key_label
        self.key_guid = key_guid

        self._index2label = {i: label for i, label in enumerate(self.labels)}
        self._label2index = {label: i for i, label in enumerate(self.labels)}

        self._keymap = {
            "text_a": self.key_text_a,
            "text_b": self.key_text_b,
            "label": self.key_label,
            "guid": self.key_guid,
        }

        self.tokenizer = self.get_tokenizer()

    def get_tokenizer(self):
        return None

    @property
    def num_labels(self):
        return len(self.labels)

    def label2index(self, label: Union[int, str]) -> Optional[int]:
        return self._label2index.get(label)

    def index2label(self, index: int) -> Optional[Union[int, str]]:
        return self._index2label.get(index)

    def encode_label(self, label: Union[int, str]) -> Optional[int]:
        return self.label2index(label)

    def decode(self, index: int) -> Union[int, str]:
        return self.index2label(index)

    def read(self, data_dir: str, **kwargs):
        return list(read_json(data_dir))


class TextProcessor(BaseSequenceClassificationProcessor):
    def __init__(
        self,
        labels: List[Union[int, str]],
        tokenizer_dir: str = "jieba.lcut",
        max_length: int = 128,
        key_text_a: str = "text_a",
        key_text_b: str = "text_b",
        key_label: str = "label",
        key_guid: str = "_id",
        concat: str = " ",
        vocab: Optional[Union[str, List[str]]] = None,
        vocab_corpus: Optional[Union[str, Iterable]] = None,
        load_dir: Optional[str] = None,
        **vocab_kwargs
    ):
        """
        Keyword Arguments:
        labels:List[Union[int,str]]        --
        vocab :str                     --
        tokenize:Callable[[str],List[str]] -- oov_subword:bool                   -- (default True)
        padding_idx:int                    -- (default 0)
        unk_idx:int                        -- (default 1)
        **kwargs                           --
        """

        super(TextProcessor, self).__init__(
            labels,
            tokenizer_dir,
            max_length,
            key_text_a,
            key_text_b,
            key_label,
            key_guid,
            concat=concat,
            **vocab_kwargs
        )

        self.concat = concat

        self.vocabulary = None
        if vocab is not None:
            self.vocabulary = Vocabulary(vocab, **vocab_kwargs)
        elif vocab_corpus is not None:
            self.vocabulary = self.build_vocabulary_from_corpus(
                vocab_corpus, **vocab_kwargs
            )

            print("built")

    def get_tokenizer(self):
        return instantiate_interface(self.tokenizer_dir)

    def build_vocabulary_from_corpus(self, corpus: Union[str, Iterable], **kwargs):
        if isinstance(corpus, str) and os.path.exists(corpus):
            corpus = list(read_json(corpus))

        texts = []
        for c in corpus:
            example = None
            if isinstance(c, dict):
                example_dict = {k: c.get(v) for k, v in self._keymap.items()}
                example = Example(**example_dict)
            elif isinstance(c, Example):
                example = c

            assert example is not None

            text_a = example.text_a if example.text_a is not None else ""
            text_b = example.text_b if example.text_b is not None else ""
            texts += [text_a, text_b]

        return Vocabulary.build_from_corpus(texts, self.tokenizer, **kwargs)

    def processing(
        self, example: Union[Example, Dict], is_train: bool = True, max_length=None
    ):
        if isinstance(example, dict):
            example_dict = {k: example.get(v) for k, v in self._keymap.items()}
            example = Example(**example_dict)

        max_length = self.max_length if max_length is None else max_length
        text_a = example.text_a if example.text_a is not None else ""
        text_b = example.text_b if example.text_b is not None else ""
        text = self.concat.join([text_a, text_b]).strip()
        token = self.tokenizer(text)
        encoded = self.vocabulary(token, max_length)

        if is_train:
            return encoded, self.encode_label(example.label)
        else:
            return encoded

    def get_dataset(self, data_dir, is_train: bool = True, **kwargs):
        data = list(self.read(data_dir, **kwargs))
        return self.data2dataset(data, is_train, **kwargs)

    def data2dataset(self, datas, is_train: bool = True, **kwargs):

        if self.vocabulary is None:
            # use default params for vocabulary
            self.vocabulary = self.build_vocabulary_from_corpus(datas)

        processed = []
        for data in tqdm(datas):
            processed += [self.processing(data, is_train=is_train, **kwargs)]

        if is_train:
            processed = sorted(processed, key=lambda x: len(x[0]))
            x, y = [*zip(*processed)]
            return SequenceClassificationDataset(x, y)
        else:
            return SequenceClassificationDataset(x=processed)

    @staticmethod
    def dataset2dataloader(dataset, batch_size, **kwargs):
        is_train = dataset.y is not None

        def collate_fn(batch):
            padding_batch = []
            for b in batch:
                if is_train:
                    x, y = b
                    padding_batch += [
                        (
                            torch.tensor(x),
                            torch.tensor(y),
                        )
                    ]
                else:
                    x = b
                    padding_batch += [torch.tensor(x)]

            if is_train:
                return default_collate(padding_batch)
            else:
                return default_collate(padding_batch), None

        kwargs.update(collate_fn=collate_fn)

        return DataLoader(dataset, batch_size, **kwargs)

    def save(self, save_dir):
        self.update_hparams()
        self.vocabulary.update_hparams()

        save_dir = os.path.abspath(save_dir)
        os.makedirs(save_dir, exist_ok=True)

        # save vocab
        vocab = self.vocabulary.hparams.pop("vocab")
        vocab_save_path = os.path.join(save_dir, "vocab.txt")
        dump_vocab_txt(vocab, vocab_save_path)
        self.vocabulary.hparams.update({"vocab": vocab_save_path})
        self.hparams.update(**self.vocabulary.hparams)

        # save labels
        labels_save_path = os.path.join(save_dir, "labels.json")
        dump_json(self.labels, labels_save_path)
        self.hparams.update({"labels": labels_save_path})

        # compatible with hydra
        self.hparams.update({"_target_": get_class_fullname(self)})

        # save processor
        processor_save_path = os.path.join(save_dir, "processor.yaml")
        dump_config(self.hparams, processor_save_path)

    @classmethod
    def load(cls, load_dir=None, **kwargs):
        """
        load from a specific dir and use kwargs to overwrite
        update vocab & labels if files exist in load_dir
        kwargs:the first priority
        """
        config = dict()
        if load_dir is not None and os.path.exists(
            os.path.join(load_dir, "processor.yaml")
        ):
            processor_save_path = os.path.join(load_dir, "processor.yaml")
            config = read_config(processor_save_path)

            # compatible with hydra
            config.pop("_target_", None)

            labels_path = config["labels"]
            labels_path = os.path.join(load_dir, os.path.basename(labels_path))
            config["labels"] = labels_path

            vocab_path = config["vocab"]
            vocab_path = os.path.join(load_dir, os.path.basename(vocab_path))
            config["vocab"] = vocab_path

            assert os.path.exists(config["labels"])
            assert os.path.exists(config["vocab"])

        config.update(**kwargs)
        return cls(**config)


class PretrainedModelProcessor(BaseSequenceClassificationProcessor):
    def __init__(
        self,
        tokenizer_dir: str,
        labels: List[Union[str, int]],
        max_length: int = 100,
        key_text_a: str = "text_a",
        key_text_b: str = "text_b",
        key_label: str = "label",
        key_guid: str = "_id",
        **tokenizer_kwargs
    ):
        """
        Keyword Arguments:
        tokenizer_dir:str             --
        labels:List[Union[str, int]]  --
        max_length:int                -- (default 100)
        key_text_a:str                -- (default 'text_a')
        key_text_b:str                -- (default 'text_b')
        key_label:str                 -- (default 'label')
        key_guid:str                  -- (default '_id')
        """

        self._tokenizer_kwargs = tokenizer_kwargs
        super(PretrainedModelProcessor, self).__init__(
            labels,
            tokenizer_dir,
            max_length,
            key_text_a,
            key_text_b,
            key_label,
            key_guid,
        )

    def get_tokenizer(self):
        return self._tokenizer_init(self.tokenizer_dir, **self._tokenizer_kwargs)

    def get_dataset(self, data_dir: str, is_train: bool = True, **kwargs):
        datas = list(self.read(data_dir, **kwargs))
        return self.data2dataset(datas, is_train, **kwargs)

    def data2dataset(
        self, datas: List[Union[Example, Dict]], is_train: bool = True, **kwargs
    ):
        # TODO: return dict-like result,use new api
        input_ids = []
        attention_mask = []
        token_type_ids = []
        labels = []
        y = None
        for data in tqdm(datas):
            processed = self.processing(data, is_train)
            input_ids += [processed["input_ids"]]
            attention_mask += [processed["attention_mask"]]
            token_type_ids += [processed["token_type_ids"]]
            labels += [processed.get("label")]

        input_ids = torch.tensor(input_ids, dtype=torch.long)
        attention_mask = torch.tensor(attention_mask, dtype=torch.long)
        token_type_ids = torch.tensor(token_type_ids, dtype=torch.long)

        x = TensorDataset(input_ids, attention_mask, token_type_ids)
        if is_train:
            y = torch.tensor(labels)
        return SequenceClassificationDataset(x, y)

    @staticmethod
    def dataset2dataloader(dataset, batch_size, **kwargs):
        is_train = dataset.y is not None

        def collate_fn(batch):
            if is_train:
                x, y = default_collate(batch)
                y = y.type(torch.long)
                return x, y
            else:
                x = default_collate(batch)
                return x, None

        kwargs.update(collate_fn=collate_fn)

        return DataLoader(dataset, batch_size, **kwargs)

    def processing(self, example: Union[Example, Dict], is_train: bool = True):
        if isinstance(example, dict):
            text_a = example.get(self.key_text_a)
            text_b = example.get(self.key_text_b)
            guid = example.get(self.key_guid)
            label = example.get(self.key_label)
            example = Example(text_a, text_b, label, guid)
        encoded = self.tokenizer(
            example.text_a,
            example.text_b,
            max_length=self.max_length,
            padding="max_length",
            truncation=True,
        )

        if is_train:
            label = self.encode_label(example.label)
            encoded.update(label=label)
        return encoded

    def save(self, save_dir: str):
        self.update_hparams()

        save_dir = os.path.abspath(save_dir)
        os.makedirs(save_dir, exist_ok=True)

        # save tokenizer
        # XXX: Toakenizer.from_pretrained(...) config.json not found
        self.tokenizer.save_pretrained(save_dir)
        self.hparams.update(tokenizer_dir=save_dir)

        # save labels
        labels_save_path = os.path.abspath(os.path.join(save_dir, "labels.json"))
        dump_json(self.labels, labels_save_path)
        self.hparams.update(labels=labels_save_path)

        # save processor
        processor_save_path = os.path.join(save_dir, "processor.yaml")
        dump_config(self.hparams, processor_save_path)

    @classmethod
    def load(cls, load_dir=None, **kwargs):
        config = dict()
        if load_dir is not None and os.path.exists(
            os.path.join(load_dir, "processor.yaml")
        ):
            processor_save_path = os.path.join(load_dir, "processor.yaml")
            config = read_config(processor_save_path)

            assert os.path.exists(config["labels"])

        config.update(**kwargs)
        return cls(**config)

    @staticmethod
    def _tokenizer_init(tokenizer_dir, backup_tokenizer=BertTokenizer, **kwargs):
        try:
            return AutoTokenizer.from_pretrained(tokenizer_dir, **kwargs)
        except Exception as e:
            return backup_tokenizer.from_pretrained(tokenizer_dir, **kwargs)
